
/*
 * LIB/ALLOC.C	- memory allocation
 *
 * These routines are used to allocate buffers with the express intent 
 * of *truely* deallocating the memory from the page table on free.
 * malloc()/free() does not normally do this.
 *
 * If you are running a large number of incoming connections or outgoing
 * feeds, using either USE_ANON_MMAP or USE_FALL_MMAP is an absolute 
 * requirement.  If you are not, then it doesn't matter.
 */

#include "defs.h"

Prototype void *pagealloc(int *psize);
Prototype void pagefree(void *p);

static int PageSize;
static int ZoneFd = -1;		/* only necessary for USE_FALL_MMAP */

#if USE_ANON_MMAP

/*
 * Our first choice is to use MAP_ANON if the system understands it
 */

void *
pagealloc(int *psize)
{
    void *b;
    if (PageSize == 0)
	PageSize = getpagesize();

    *psize = PageSize;
    b = (void *)mmap((caddr_t)0, PageSize, PROT_READ|PROT_WRITE, MAP_ANON, -1, 0);
    if (b == NULL || b == (void *)-1) {
	syslog(LOG_EMERG, "mmap/allocation failed: %s", strerror(errno));
	exit(1);
    }
    return(b);
}

void
pagefree(void *page)
{
    if (page != NULL) {
	munmap((caddr_t)page, PageSize);
    }
}

#else
#if USE_FALL_MMAP

/*
 * Our second choice is to use a MAP_PRIVATE file mmap()
 */

void *
pagealloc(int *psize)
{
    void *b;

    if (PageSize == 0)
	PageSize = getpagesize();

    if (ZoneFd < 0) {
	int i;
	int pid = (int)getpid();

	for (i = 0; i < 1000; ++i) {
	    char path[256];

	    sprintf(path, "/tmp/diablo.zone.%d.%d", pid, i);
	    if ((ZoneFd=open(path, O_CREAT|O_TRUNC|O_RDWR|O_EXCL, 0600)) >= 0){
		remove(path);
		break;
	    }
	}
	if (ZoneFd < 0) {
	    syslog(LOG_EMERG, "unable to create %dK zone file in /tmp: %s", 
		PageSize / 1024,
		strerror(errno)
	    );
	    exit(1);
	}
	if (ftruncate(ZoneFd, PageSize) < 0) {
	    syslog(LOG_EMERG, "unable to extend zone file in /tmp: %s", 
		strerror(errno)
	    );
	    exit(1);
	}
	fcntl(ZoneFd, F_SETFD, 1);	/* set close-on-exec */
    }

    *psize = PageSize;
    b = (void *)mmap((caddr_t)0, PageSize, PROT_READ|PROT_WRITE, MAP_PRIVATE, ZoneFd, 0);
    if (b == NULL || b == (void *)-1) {
	syslog(LOG_EMERG, "mmap/allocation failed: %s", strerror(errno));
	exit(1);
    }
    return(b);
}

void
pagefree(void *page)
{
    if (page != NULL) {
	munmap((caddr_t)page, PageSize);
    }
}

#else

void *
pagealloc(int *psize)
{
    void *b;

    if (PageSize == 0)
	PageSize = getpagesize();

    *psize = PageSize;
    b = malloc(PageSize);
    if (b == NULL) {
	syslog(LOG_EMERG, "mmap/allocation failed: %s", strerror(errno));
	exit(1);
    }
    return(b);
}

void
pagefree(void *page)
{
    if (page != NULL) {
	free(page);
    }
}

#endif
#endif

