
/*
 * DCLEAN.C	Scan diablo spool directories and remove articles not 
 *		found in the dhistory file or articles marked as 
 *		expired in the dhistory file.
 *
 *	This is a general 'clean up any problems' program.  Since it
 *	does not actually have to open any article files, it is extremely
 *	efficient.  This program is commonly run after a crash.  Note
 *	that due to the way they are named, it makes no sense to leave
 *	files around that do not exist in the dhistory file.
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"
#if USE_SYSV_DIR
#include <dirent.h>
#else
#include <sys/dir.h>
#endif

void ScanDirectory(const char *dname);

int RemoveOpt;
int VerboseOpt = 1;	/* verbose by default	*/
int QuietOpt;
int Count = 0;

int
main(int ac, char **av)
{
    DIR *dir;
    char *dirName = NULL;
    int i;

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];

	if (*ptr == '-') {
	    ptr += 2;
	    switch(ptr[-1]) {
	    case 'q':
		QuietOpt = 1;
		if (*ptr)
		    QuietOpt = strtol(ptr, NULL, 0);
		VerboseOpt = 0;
		break;
	    case 'v':
		VerboseOpt = 1;
		QuietOpt = 0;
		break;
	    case 'r':
		RemoveOpt = 1;
		break;
	    default:
		printf("illegal option: %s\n", ptr - 2);
		exit(1);
	    }
	} else if (dirName == NULL) {
	    dirName = ptr;
	} else {
	    printf("illegal argument: %s\n", ptr);
	    exit(1);
	}
    }
    if (ac == 1) {
	puts("dclean [-remove] [-quiet] [-verbose] [directory]");
	exit(0);
    }

    if (QuietOpt < 2)
	printf("DCLEAN running\n");

    /*
     * Open the history file
     */

    HistoryOpen(NULL, 0);

    /*
     * Scan a specific directory, or scan the entire news spool
     */

    if (dirName) {
	if (chdir(dirName) < 0) {
	    perror("chdir");
	    exit(1);
	}
	ScanDirectory(dirName);
    } else {
	if ((dir = opendir(SpoolHome)) != NULL) {
#if USE_SYSV_DIR
	    struct dirent *den;
#else
	    struct direct *den;
#endif

	    while ((den = readdir(dir)) != NULL) {
		if (strncmp(den->d_name, "D.", 2) == 0) {
		    long gmt = time(NULL) / 60;
		    long dgmt;

		    if (sscanf(den->d_name + 2, "%lx", &dgmt) != 1) {
			fprintf(stderr, "malformed directory name: %s\n", den->d_name);
			continue;
		    }

		    /*
		     * skip directories that are too close to now, in case
		     * there are any article files 'in-progress' that have
		     * not yet been entered into the dhistory database.
		     */
		    dgmt -= dgmt % 10;
		    gmt -= gmt % 10;
		    if (dgmt == gmt || dgmt == gmt - 10) {
			if (VerboseOpt)
			    fprintf(stderr, "%s too close to now, skipping\n", den->d_name);
			continue;
		    }

		    if (chdir(SpoolHome) < 0) {
			perror("chdir");
			exit(1);
		    }
		    if (chdir(den->d_name) < 0) {
			fprintf(stderr, "Unable to chdir to %s/%s, skipping\n", 
			    SpoolHome, 
			    den->d_name
			);
		    } else {
			ScanDirectory(den->d_name);
		    }
		}
	    }
	    closedir(dir);
	} else {
	    fprintf(stderr, "unable to scan %s\n", SpoolHome);
	}
    }
    if (QuietOpt < 2)
	printf("DCLEAN end, %d files removed\n", Count);
    return(0);
}

/*
 * DIABLO uses a two-level directory tree.  Therefore we
 * do not have to worry about recursion.
 */

void
ScanDirectory(const char *dname)
{
    DIR *dir = opendir(".");

    if (dir != NULL) {
#if USE_SYSV_DIR
	struct dirent *den;
#else
	struct direct *den;
#endif
	hash_t hv;
	int iter;
	Node *base = NULL;

	/*
	 * we pre-scan the directory in order to avoid
	 * problems with hashed filesystems whereby the
	 * directory scan misses some files when you
	 * have embedded remove()'s or file creates.
	 */

	{
	    Node **lnext = &base;

	    while ((den = readdir(dir)) != NULL) {
		Node *node = MakeNode(den->d_name, 0);
		*lnext = node;
		lnext = &node->no_Next;
	    }
	}

	closedir(dir);

	/*
	 * The hash_t values are embedded in the file name, allowing
	 * us to look them up in the history database without having
	 * to open the file.
	 */

	while (base) {
	    Node *node = base;

	    base = node->no_Next;

	    if (sscanf(node->no_Name, "%lx.%lx.%x", &hv.h1, &hv.h2,&iter) == 3) {
		History h;

		if (HistoryLookupByHash(hv, &h) == 0) {
		    /*
		     * Check for files with unused iteration numbers (which can occur
		     * when two or more diablo forks receive the same article message-id
		     * at the same time), and for files which were supposed to have
		     * been expired.
		     */
		    if (h.iter != (unsigned short)iter || h.exp == (unsigned short)-1) {
			if (VerboseOpt) {
			    if (h.iter == (unsigned short)iter)
				printf("EXPIR %s/%s", dname, node->no_Name);
			    else
				printf("ITER%d %s/%s", h.iter, dname, node->no_Name);
			}
			if (RemoveOpt) {
			    if (remove(node->no_Name) == 0) {
				++Count;
				if (VerboseOpt)
				    printf(" (removed)");
			    }
			}
			if (VerboseOpt)
			    printf("\n");
		    }
		} else {
		    /*
		     * Or, if there is nothing in the history file at all,
		     * just delete the file.
		     */
		    if (VerboseOpt)
			printf("NOEXIST %s/%s", dname, node->no_Name);
		    if (RemoveOpt) {
			if (remove(node->no_Name) == 0) {
			    ++Count;
			    if (VerboseOpt)
				printf(" (removed)");
			}
		    }
		    if (VerboseOpt)
			printf("\n");
		}
	    }
	    FreeNode(node);
	}
    }
}

