
/*
 * NEWSLINK.C
 *
 * newslink -b batchfile -h host -t timeout -l log-after 
 * 	-c close-reopen-after -pipe ...other options
 *
 * (c)Copyright 1996, Matthew Dillon.  Permission granted for global
 * redistribution as long as attributations are left intact.
 *
 * This program will attempt to run one or more batch files containing
 * article references of the form:
 *
 * Relative-Path <message-id>
 * Relative-Path <message-id>
 *
 * For example:
 *
 * de/comm/chatsystems/911 <4mle9s$mef@nz12.rz.uni-karlsruhe.de>
 * alt/binaries/multimedia/d/3643 <4mme9r$plc@nntp1.best.com>
 * 
 * The primary methodology of using this program is to have another
 * program, SPOOLOUT, generate and maintain numerically indexed spool files.
 * That program forks and runs this program to actually process the spool
 * files.
 *
 * This program's job is to process one or more spool files (performing locking
 * and skipping locked spool files), spool the articles in question to the
 * remote hosts, then delete the spool files as process is completed, or
 * rewrite the spool files if a failure occurs.
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"

#define DIABLO

#define T_ACCEPTED	1
#define T_REFUSED	2
#define T_REJECTED	3
#define T_FAILED	4
#define T_STREAMING	5
#define T_FAILEDEXIT	6

#define INF_HELP        100     /* Help text on way */
#define INF_AUTH        180     /* Authorization capabilities */
#define INF_DEBUG       199     /* Debug output */

#ifndef SPOOL
#define SPOOL		"/news/spool/news"
#endif
#ifndef OUTGOING
#ifdef DIABLO
#define OUTGOING	"/news/dqueue"
#else
#define OUTGOING	"/news/spool/out.going"
#endif
#endif

#define MAXCLINE	8192
#define MAXFILEDES	32

#define OUR_DELAY	1
#define THEIR_DELAY	2
#define DDTIME		30

#define OK_CANPOST      200     /* Hello; you can post */
#define OK_NOPOST       201     /* Hello; you can't post */
#define OK_SLAVE        202     /* Slave status noted */
#define OK_STREAMOK	203	/* Can-do streaming	*/
#define OK_GOODBYE      205     /* Closing connection */
#define OK_GROUP        211     /* Group selected */
#define OK_GROUPS       215     /* Newsgroups follow */
#define OK_ARTICLE      220     /* Article (head & body) follows */
#define OK_HEAD         221     /* Head follows */
#define OK_BODY         222     /* Body follows */
#define OK_NOTEXT       223     /* No text sent -- stat, next, last */
#define OK_NEWNEWS      230     /* New articles by message-id follow */
#define OK_NEWGROUPS    231     /* New newsgroups follow */
#define OK_XFERED       235     /* Article transferred successfully */
#define OK_STRMCHECK	238	/* check response / want article	*/
#define OK_STRMTAKE	239	/* takeit response / article received	*/
#define OK_POSTED       240     /* Article posted successfully */
#define OK_AUTHSYS      280     /* Authorization system ok */
#define OK_AUTH         281     /* Authorization (user/pass) ok */

#define CONT_XFER       335     /* Continue to send article */
#define CONT_POST       340     /* Continue to post article */
#define NEED_AUTHINFO   380     /* authorization is required */
#define NEED_AUTHDATA   381     /* <type> authorization data required */

#define ERR_GOODBYE     400     /* Have to hang up for some reason */
#define ERR_NOGROUP     411     /* No such newsgroup */
#define ERR_NCING       412     /* Not currently in newsgroup */
#define ERR_NOCRNT      420     /* No current article selected */
#define ERR_NONEXT      421     /* No next article in this group */
#define ERR_NOPREV      422     /* No previous article in this group */
#define ERR_NOARTIG     423     /* No such article in this group */
#define ERR_NOART       430     /* No such article at all */
#define ERR_RESEND	431	/* please resend the article	*/
#define ERR_GOTIT       435     /* Already got that article, don't send */
#define ERR_XFERFAIL    436     /* Transfer failed */
#define ERR_XFERRJCT    437     /* ihave, Article rejected, don't resend */
#define ERR_STRMCHECK	438	/* check response / do not want article */
#define ERR_STRMTAKE	439	/* takeit response / article failed	*/
#define ERR_NOPOST      440     /* Posting not allowed */
#define ERR_POSTFAIL    441     /* Posting failed */
#define ERR_NOAUTH      480     /* authorization required for command */
#define ERR_AUTHSYS     481     /* Authorization system invalid */
#define ERR_AUTHREJ     482     /* Authorization data rejected */

#define ERR_COMMAND     500     /* Command not recognized */
#define ERR_CMDSYN      501     /* Command syntax error */
#define ERR_ACCESS      502     /* Access to server denied */
#define ERR_FAULT       503     /* Program fault, command not performed */
#define ERR_AUTHBAD     580     /* Authorization Failed */

#define ERR_UNKNOWN	990

/*
 * Streaming parameters
 */

#define STREAM_OFF	0
#define STREAM_RELOAD	1
#define STREAM_ON	2

#define STATE_EMPTY	0	/* empty slot				*/
#define STATE_CHECK	1	/* check transmitted			*/
#define STATE_POSTED	2	/* takethis + article transmitted	*/

#define MAXSTREAM	16	/* Minimum is 2				*/
#define MAXPENDBYTES	1024 - (MAXSTREAM * 8)	/* maximum pending bytes*/

typedef struct Stream {
    int	st_State;		/* state	*/
    int st_DumpRCode;
    char *st_RelPath;		/* path		*/
    char *st_MsgId;		/* message id	*/
} Stream;

int connectTo(const char *hostName, const char *serviceName, int defPort);
int Transact(int cfd, const char *relPath, const char *msgId);
int DumpArticle(int cfd, const char *relPath);
int StreamTransact(int cfd, const char *relPath, const char *msgId);
void StreamReload(int cfd);
Stream *LocateStream(const char *msgId, int state);
void RefilePendingStreams(FILE *fo);
int commandResponse(int cfd, char **rptr, const char *ctl, ...);
void clearResponseBuf(void);
int readline(int fd, char *buf, int size);
void logit(const char *ctl, ...);
void logStats(const char *description);
void credtime(int whos);
void credreset(void);
int ValidMsgId(char *msgid);

int LogAfter = 1000;
int CloseReopenAfter = 1000;
int Timeout = 600;
int DebugOpt = 0;
int DeleteDetectOpt = 0;
int WaitTime = 10;

Stream StreamAry[MAXSTREAM];
int StreamMode = 0;		/* set by stream negotiation	*/
int StreamPend = 0;		/* pending streaming requests	*/
int BytesPend = 0;

int LogAfterCount = -1;
int CloseReopenCount = -1;
int KillFd = -1;

char *HostName;
char *BatchFileCtl;
char CurrentBatchFile[1024];
char LastErrBuf[256];
int BatchSeq = -1;
int NumBatches = 1;

int AcceptedTotal;
int RefusedTotal;
int RejectedTotal;
int CountTotal;
int PipeOpt;
time_t TimeStart;

int AcceptedCnt;
int RefusedCnt;
int RejectedCnt;
int Count;
time_t DeltaStart;

int TermFlag;
int StatSize;
char *StatBuf;

int OurMs;
int TheirMs;
int MsCount;

int OurMsTotal;
int TheirMsTotal;
int MsCountTotal;

void sigTerm(int sigNo);
void bsprintf(const char *ctl, ...);

int
main(int ac, char **av)
{
    int i;
    int cfd;
    int fd;
    int eNoBat = 0;

    openlog("newslink", LOG_PERROR | LOG_NDELAY | LOG_PID, LOG_NEWS);

    if (chdir(OUTGOING) != 0) {
	logit("couldn't chdir to %s\n", OUTGOING);
	exit(1);
    }

    for (i = 1; i < ac; ++i) {
	char *ptr = av[i];

	ptr += 2;
	switch(ptr[-1]) {
	case 's':
	    ptr = (*ptr) ? ptr : av[++i];
	    if (strlen(ptr) > 23) {
		StatBuf = ptr;
		StatSize = strlen(ptr);
	    }
	    break;
	case 'D':
	    DeleteDetectOpt = 1;
	    break;
	case 'p':
	    PipeOpt = 1;
	    break;
	case 'd':
	    DebugOpt = (*ptr) ? strtol(ptr, NULL, 0) : 1;
	    break;
	case 'b':
	    BatchFileCtl = (*ptr) ? ptr : av[++i];
	    break;
	case 'S':
	    BatchSeq = strtol(((*ptr) ? ptr : av[++i]), NULL, 0);
	    break;
	case 'N':
	    NumBatches = strtol(((*ptr) ? ptr : av[++i]), NULL, 0);
	    break;
	case 'h':
	    HostName = (*ptr) ? ptr : av[++i];
	    break;
	case 'w':
	    WaitTime = strtol(((*ptr) ? ptr : av[++i]), NULL, 0);
	    break;
	case 't':
	    Timeout = strtol(((*ptr) ? ptr : av[++i]), NULL, 0);
	    break;
	case 'l':
	    LogAfter = strtol(((*ptr) ? ptr : av[++i]), NULL, 0);
	    break;
	case 'c':
	    CloseReopenAfter = strtol(((*ptr) ? ptr : av[++i]), NULL, 0);
	    break;
	}
    }
    LogAfterCount += LogAfter;
    CloseReopenCount += CloseReopenAfter;

    if (BatchFileCtl == NULL || HostName == NULL || i > ac) {
	printf("newslink %s\n", VERSION);
	puts(av[0]);
	puts("    -b batchfile          - specify batchfile");
	puts("	  -b template%d		- template containing %[xx]d");
	puts("    -h host               - specify remote INND host");
	puts("    -t #                  - specify timeout, default 600s");
	puts("    -l #                  - log-after-count, default 1000");
	puts("	  -c #                  - close-reopen-after, default 1000");
	puts("	  -s \"<24-spaces>\"    - /bin/ps argv space for status");
	puts("    -D                    - detect delete-out-from-under");
	puts("	  -p                    - input on stdin rather then file");
	puts("	  -d[#]			- debug option");
	puts("	  -S #			- starting sequence #, template mode");
	puts("	  -N #                  - number of batchfiles to process, template mode");
	puts("    -w #                  - reconnect-after-failure delay");
	exit(1);
    }

    rsignal(SIGPIPE, SIG_IGN);
    rsignal(SIGHUP, sigTerm);
    rsignal(SIGINT, sigTerm);
    rsignal(SIGALRM, sigTerm);

    if (*BatchFileCtl == 0) {
	logit("batchfile is null!\n");
	exit(0);
    }

    while (--NumBatches >= 0) {
	time_t ddTime;

	AcceptedTotal = 0;
	RefusedTotal = 0;
	RejectedTotal = 0;
	CountTotal = 0;

	AcceptedCnt = 0;
	RefusedCnt = 0;
	RejectedCnt = 0;
	Count = 0;

	OurMs = 0;
	TheirMs = 0;
	MsCount = 0;

	OurMsTotal = 0;
	TheirMsTotal = 0;
	MsCountTotal = 0;

	if (BatchSeq < 0) {
	    sprintf(CurrentBatchFile, "%s", BatchFileCtl);
	} else {
	    sprintf(CurrentBatchFile, BatchFileCtl, BatchSeq++);
	}

	/*
	 * Open next batchfile
	 */

	if (PipeOpt) {
	    fd = 0;
	} else {
	    fd = open(CurrentBatchFile, O_RDWR);
	    if (fd < 0) {
		if (eNoBat == 0) {
		    eNoBat = 1;
		    logit("no batchfile\n");
		}
		continue;
	    }

	    /*
	     * Lock batchfile
	     */
	    {
		struct stat st;

		if (xflock(fd, XLOCK_EX|XLOCK_NB) != 0) {
		    if (eNoBat == 0) {
			eNoBat = 1;
			logit("batchfile already locked\n");
		    }
		    close(fd);
		    continue;
		}
		if (fstat(fd, &st) != 0) {
		    logit("fstat failed: %s\n", strerror(errno));
		    exit(1);
		}
		if (st.st_nlink == 0) {
		    close(fd);
		    continue;
		}
		eNoBat = 0;
	    }
	    fprintf(stderr, "%s\n", CurrentBatchFile);
	}

	TimeStart = DeltaStart = ddTime = time(NULL);

	/*
	 * Connect to remote, send news
	 */

	if ((cfd = connectTo(HostName, "nntp", 119)) >= 0) {
	    char buf[4096];
	    char relPath[4096];
	    char msgId[4096];
	    char path[1024];
	    FILE *fo = NULL;
	    int connectCount = 0;
	    int bufInval = 1;
	    int loggedMark = 0;

	    credtime(0);

	    if (PipeOpt)
		sprintf(path, "%s", CurrentBatchFile);
	    else
		sprintf(path, "%s.tmp", CurrentBatchFile);

	    bsprintf("wait/in %d", connectCount);

	    /*
	     * Loop if need to reconnect, there are pending streaming commands,
	     * or we can get a valid buffer.  Note that this allows us to halt
	     * readline() calls when we reach our streaming limit as well as to
	     * do a final-drain when we reach the end of the buffer.
	     */

	    for (;;) {
		/*
		 * If there is no active buffer and we have not reached
		 * our streaming limit, read in another article control line.
		 */
		if (bufInval && StreamPend < MAXSTREAM && BytesPend < MAXPENDBYTES)
		    bufInval = readline(fd, buf, sizeof(buf));

		/*
		 * If there is nothing left to do, break out of the loop
		 */

		if (bufInval && StreamPend == 0)
		    break;

		/*
		 * Go.
		 */

		if (DebugOpt)
		    printf("bufInval %d StreamPend %d (%s)\n", bufInval, StreamPend, (bufInval) ? "?" : buf);

		if (DeleteDetectOpt && 
		    !PipeOpt && 
		    (long)(time(NULL) - ddTime) > DDTIME
		) {
		    struct stat st;

		    if (stat(CurrentBatchFile, &st) != 0)
			break;
		    ddTime = time(NULL);
		}

		if (bufInval || (sscanf(buf, "%s %s", relPath, msgId) == 2 && ValidMsgId(msgId))) {
		    int t;

		    bsprintf("process %d", connectCount);

		    if (bufInval)
			t = Transact(cfd, NULL, NULL);
		    else
			t = Transact(cfd, relPath, msgId);

		    if (DebugOpt)
			printf("Transaction result: %d %s", t, (bufInval) ? "<inval>\n" : buf);

		    bsprintf("wait/in %d", connectCount);

		    switch(t) {
		    case T_STREAMING:
			/*
			 * operation in progress, will return the real status
			 * later.
			 */
			break;
		    case T_ACCEPTED:
			++AcceptedCnt;
			++AcceptedTotal;
			++connectCount;
			break;
		    case T_REFUSED:
			++RefusedCnt;
			++RefusedTotal;
			++connectCount;
			break;
		    case T_REJECTED:
			++RejectedCnt;
			++RejectedTotal;
			++connectCount;
			break;
		    case T_FAILED:
			/*
			 * If failed, attempt to reconnect to remote.
			 */
			break;
		    }

		    /*
		     * Many INND's are setup to close a connection after a
		     * certain period of time.  Thus, a failure could be a
		     * normal occurance.  If we have already had at least one
		     * successfull transaction, we attempt to reconnect when
		     * this case occurs.
		     */

		    if (t == T_FAILED || t == T_FAILEDEXIT) {
			if (connectCount > 0 && t != T_FAILEDEXIT && TermFlag == 0) {
			    logit("Remote EOF, attempting to reconnect\n");
			    logStats("mark");
			    loggedMark = 1;
			    close(cfd);
			    credtime(OUR_DELAY);
			    if ((cfd = connectTo(HostName, "nntp", 119)) >= 0) {
				connectCount = 0;
				if (CloseReopenCount > 0)
				    CloseReopenCount = CountTotal+CloseReopenAfter;
				credtime(0);
				continue;
			    }
			    credtime(0);
			}
			NumBatches = 0;
			if (fo == NULL)
			    fo = fopen(path, "w");
			if (fo != NULL) {
			    RefilePendingStreams(fo);
			    if (!bufInval)
				fputs(buf, fo);
			}
			break;
		    }

		    /*
		     * buffer successfully dealt with, buffer is now invalid
		     */

		    bufInval = 1;

		    /*
		     * Bump transaction count
		     */

		    if (t != T_STREAMING) {
			++Count;
			++CountTotal;
		    }

		    /*
		     * Log stats after specified count
		     */

		    if (LogAfterCount > 0 && CountTotal >= LogAfterCount) {
			loggedMark = 1;
			logStats("mark");
			LogAfterCount += LogAfter;
		    }

		    /*
		     * Close remote to allow remote logging to occur after
		     * specified count (doesn't work with streaming yet)
		     */

		    if (
			StreamMode == STREAM_OFF &&
			CloseReopenCount > 0 && 
			CountTotal >= CloseReopenCount
		    ) {
			close(cfd);
			credtime(OUR_DELAY);
			if ((cfd = connectTo(HostName, "nntp", 119)) < 0) {
			    credtime(0);
			    break;
			}
			credtime(0);
			connectCount = 0;
			CloseReopenCount += CloseReopenAfter;
		    }
		} else {
		    logit("Buffer syntax error: %s\n", buf);
		    bufInval = 1;
		}

		/*
		 * If a signal occured during the transaction, break out of
		 * our loop.
		 */

		if (TermFlag) {
		    logit("Terminated with signal\n");
		    break;
		}
	    }

	    /*
	     * Final delta stats, only if we had
	     * previously logged marks, otherwise the
	     * mark stats will be the same as the final
	     * stats and we do not bother logging the mark.
	     */

	    if (loggedMark != 0)
		logStats("mark");

	    /*
	     * Rewrite batchfile
	     */
	    
	    if (StreamPend) {
		if (fo == NULL)
		    fo = fopen(path, "w");
		if (fo)
		    RefilePendingStreams(fo);
	    }

	    if (readline(fd, buf, sizeof(buf)) == 0 || fo != NULL) {
		struct stat st;

		if (fo == NULL)
		    fo = fopen(path, "w");
		if (fo != NULL) {
		    do {
			fputs(buf, fo);
		    } while (readline(fd, buf, sizeof(buf)) == 0);
		    fflush(fo);
		    fclose(fo);
		    fo = NULL;
		}
		if (DeleteDetectOpt && 
		    !PipeOpt && 
		    stat(CurrentBatchFile, &st) != 0
		) {
		    remove(path);
		} else if (PipeOpt == 0) {
		    rename(path, CurrentBatchFile);
		}
	    } else {
		if (PipeOpt == 0)
		    remove(CurrentBatchFile);
	    }

	    /*
	     * Final overall stats
	     */

	    AcceptedCnt = AcceptedTotal;
	    RefusedCnt = RefusedTotal;
	    RejectedCnt = RejectedTotal;
	    Count = CountTotal;
	    DeltaStart = TimeStart;
	    OurMs += OurMsTotal;
	    TheirMs += TheirMsTotal;
	    MsCount += MsCountTotal;

	    logStats("final");

	    if (cfd >= 0)
		close(cfd);
	}
	if (fd > 0)
	    close(fd);
	if (cfd < 0)
	    break;
    }
    exit(0);
}

int
connectTo(const char *hostName, const char *serviceName, int defPort)
{
    struct hostent *host = gethostbyname(hostName);
    struct servent *serv = getservbyname(serviceName, "tcp");
    struct sockaddr_in sin;
    int fd = socket(AF_INET, SOCK_STREAM, 0);
    int cfd = -1;
    static int ConnectCount;

    clearResponseBuf();

    memset(&sin, 0, sizeof(sin));
    ++ConnectCount;

    bsprintf("connect: %d", ConnectCount);

    if (host) {
	sin.sin_family = host->h_addrtype;
	memmove(&sin.sin_addr, host->h_addr, host->h_length);
    } else if (strtol(hostName, NULL, 0) != 0) {
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = inet_addr(hostName);
    } else {
	logit("hostname lookup failure: %s\n", strerror(errno));
	fd = -1;
    }
    if (serv) {
	sin.sin_port = serv->s_port;
    } else {
	sin.sin_port = htons(defPort);
	logit("unable to lookup service '%s', using port %d\n", 
	    serviceName, 
	    defPort
	);
    }
    if (fd >= 0 && connect(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	logit("connect: %s\n", strerror(errno));
	close(fd);
	fd = -1;
	bsprintf("connect: fail");
    }

    LastErrBuf[0] = 0;

    if (fd >= 0) {
	char *ptr;

	cfd = fd;

	switch(commandResponse(cfd, &ptr, NULL)) {
	case OK_CANPOST:
	    break;
	default:
	    logit("connect: %s\n", ptr);
	    bsprintf("connect: err resp");
	    close(cfd);
	    cfd = -1;
	    fd = -1;
	    break;
	}
	credreset();
    }

    /*
     * stream ?
     *
     * If so, set the state to STREAM_RELOAD to cause any
     * queued streaming requests to be resent.
     */

    if (cfd >= 0) {
	char *ptr;

	switch(commandResponse(cfd, &ptr, "mode stream\r\n")) {
	case OK_STREAMOK:
	    StreamMode = STREAM_RELOAD;
	    logit("connect: %s (streaming)\n", ptr);
	    bsprintf("connect: stream");
	    break;
	default:
	    StreamMode = STREAM_OFF;
	    logit("connect: %s (nostreaming)\n", ptr);
	    bsprintf("connect: nostream");
	    break;
	}
    }

    /*
     * clean up
     */

    if (cfd < 0) {
	sprintf(LastErrBuf, "connect-timeout");
	sleep(WaitTime);
    }
    KillFd = fd;
    return(cfd);
}

/*
 * Transact() - begin a streaming or non-streaming transaction, or
 *		complete a streaming transaction.
 */

int
Transact(int cfd, const char *relPath, const char *msgId)
{
    int r = 0;
    char *ptr = NULL;

    /*
     * Handle streaming.  If r returns zero, we revert to the
     * old operation (this is only occurs if streaming is disabled).
     */

    if (StreamMode == STREAM_RELOAD) {
	StreamReload(cfd);
	StreamMode = STREAM_ON;
    }
    if (StreamMode == STREAM_ON) {
	r = StreamTransact(cfd, relPath, msgId);
	if (r)
	    return(r);
    }

    if (relPath == NULL)
	return(T_STREAMING);

    switch(commandResponse(cfd, &ptr, "ihave %s\r\n", msgId)) {
    case CONT_XFER:
	break;
    case ERR_XFERRJCT:
	r = T_REJECTED;
	break;
    case ERR_GOTIT:
	r = T_REFUSED;
	break;
    case ERR_ACCESS:
    case ERR_FAULT:
    case ERR_AUTHBAD:
	r = T_FAILEDEXIT;
	break;
    case ERR_NOAUTH:
    default:
	r = T_FAILED;
	break;
    }
    if (r == 0) {
	r = DumpArticle(cfd, relPath);
	switch(commandResponse(cfd, &ptr, NULL)) {
	case OK_XFERED:
	    if (r == 0)
		r = T_ACCEPTED;
	    break;
	case ERR_GOTIT:
	    if (r == 0)
		r = T_REFUSED;
	    break;
	case ERR_XFERRJCT:
	    if (r == 0)
		r = T_REJECTED;
	    break;
	case ERR_XFERFAIL:
	default:
	    if (r == 0)
		r = T_FAILED;
	    break;
	}

	/*
	 * If we couldn't send the article, we ignore the
	 * response code because the remote might send the
	 * wrong response to the null article.
	 */
	if (r < 0)
	    r = T_REJECTED;
    }
    return(r);
}

int
DumpArticle(int cfd, const char *relPath)
{
    char path[4096];
    FILE *fi = NULL;
    FILE *fo = NULL;
    int r = -1;
    int ccfd = dup(cfd);

    sprintf(path, "%s/%s", SPOOL, relPath);

    if (cfd >= 0 &&
	(fi = fopen(path, "r")) != NULL &&
	(fo = fdopen(ccfd, "w")) != NULL
    ) {
	int lastnl = 1;

	r = 0;

	if (DebugOpt) {
	    struct stat st;
	    fstat(fileno(fi), &st);
	    printf(">> (article, %d bytes)\n", (int)st.st_size);
	}

	/*
	 * In terms of credtime, we assume that our reading
	 * of the file is instantanious.  I don't want to
	 * call gettimeofday() for each line!
	 */

	credtime(OUR_DELAY);

	while (fgets(path, sizeof(path) - 1, fi) != NULL) {
	    int l = strlen(path);

	    if (lastnl && path[0] == '.')
		fputc('.', fo);

	    if (l > 0 && path[l-1] == '\n') {
		lastnl = 1;
		path[l-1] = '\r';
		path[l] = '\n';
		path[l+1] = 0;
	    } else {
		lastnl = 0;
	    }
	    fputs(path, fo);
	}
	if (lastnl == 0)
	    fprintf(fo, "\r\n");
	fflush(fo);
	credtime(THEIR_DELAY);
    } else {
	if (DebugOpt)
	    printf(">> (file not found or other error)\n");
    }

    if (fi)
	fclose(fi);

    if (fo)
	fclose(fo);
    else if (ccfd >= 0)
	close(ccfd);

    commandResponse(cfd, NULL, ".\r\n");

    return(r);
}
	
/*
 * Handle streaming protocol
 *
 * StreamReload() - retransmit pending stream after disconnect/reconnect
 * StreamTransact() - normal stream state machine
 */

void
StreamReload(int cfd)
{
    int i;

    for (i = 0; i < MAXSTREAM; ++i) {
	Stream *s = &StreamAry[i];

	if (s->st_State != STATE_EMPTY) {
	    commandResponse(cfd, NULL, "check %s\r\n", s->st_MsgId);
	    s->st_State = STATE_CHECK;
	}
    }
}

int
StreamTransact(int cfd, const char *relPath, const char *msgId)
{
    Stream *s;
    int r = T_STREAMING;

    /*
     * Another article to stream?
     */

    if (relPath) {
	if ((s = LocateStream(msgId, 0)) != NULL) {
	    return(T_REFUSED);
	}
	s = &StreamAry[StreamPend];
	s->st_MsgId  = strcpy(malloc(strlen(msgId) + 1), msgId);
	s->st_RelPath   = strcpy(malloc(strlen(relPath) + 1), relPath);
	commandResponse(cfd, NULL, "check %s\r\n", s->st_MsgId);
	s->st_State = STATE_CHECK;
	BytesPend += strlen(s->st_MsgId);
	++StreamPend;
    }

    /*
     * Should we wait for a response ?
     *
     * note: code has been written to allow for non-blocking reads in the future.
     * For now, we simply enforce
     *
     * check command response:		OK_STRMCHECK, ERR_STRMCHECK, ERR_RESEND
     * takethis command response:	OK_STRMTAKE,  ERR_STRMTAKE, ERR_RESEND
     * misc:				ERR_GOODBYE
     */

    if ((relPath == NULL && StreamPend > 0) || StreamPend > MAXSTREAM * 2 / 3 - 1) {
	char *ptr;
	Stream *s = NULL;
	int delMe = 0;

	switch(commandResponse(cfd, &ptr, NULL)) {
	case OK_STRMCHECK:
	    if ((s = LocateStream(ptr, STATE_CHECK)) != NULL) {
		int r;

		commandResponse(cfd, NULL, "takethis %s\r\n", s->st_MsgId);
		r = DumpArticle(cfd, s->st_RelPath);
		s->st_State = STATE_POSTED;
		s->st_DumpRCode = r;
	    }
	    break;
	case ERR_STRMCHECK:
	    if ((s = LocateStream(ptr, STATE_CHECK)) != NULL) {
		delMe = 1;
		r = T_REFUSED;
	    }
	    break;
	case OK_STRMTAKE:
	    if ((s = LocateStream(ptr, STATE_POSTED)) != NULL) {
		delMe = 1;
		r = T_ACCEPTED;
		if (s->st_DumpRCode < 0)
		   r = T_REJECTED;
	    }
	    break;
	case ERR_STRMTAKE:
	    if ((s = LocateStream(ptr, STATE_POSTED)) != NULL) {
		delMe = 1;
		r = T_REFUSED;
		if (s->st_DumpRCode < 0)
		   r = T_REJECTED;
	    }
	    break;
	case ERR_RESEND:
	    if ((s = LocateStream(ptr, STATE_POSTED)) != NULL) {
		delMe = 1;
		r = T_REJECTED;	/* XXX */
		if (s->st_DumpRCode < 0)
		   r = T_REJECTED;
	    }
	    break;
	case ERR_GOODBYE:
	    r = T_FAILED;
	    break;
	default:
	    r = T_FAILED;
	    break;
	}
	if (delMe) {
	    --StreamPend;
	    BytesPend -= strlen(s->st_MsgId);
	    free(s->st_RelPath);
	    free(s->st_MsgId);
	    if (s != &StreamAry[StreamPend]) {
		*s = StreamAry[StreamPend];
		memset(&StreamAry[StreamPend], 0, sizeof(Stream));
	    } else {
		memset(s, 0, sizeof(Stream));
	    }
	}
    }
    if (DebugOpt)
	printf("StreamTransact: return %d\n", r);
    return(r);
}

/*
 * Locate id
 */

Stream *
LocateStream(const char *msgId, int state)
{
    int i;
    int idLen;

    while (*msgId && *msgId != '<')
	++msgId;
    for (idLen = 0; msgId[idLen] && msgId[idLen] != '>'; ++idLen)
	;
    if (msgId[idLen] == '>')
	++idLen;

    if (DebugOpt)
	printf("MsgId(%*.*s,%d):", idLen, idLen, msgId, state);

    for (i = 0; i < MAXSTREAM; ++i) {
	Stream *s = &StreamAry[i];

	if ((state == 0 || s->st_State == state) && 
	    s->st_State &&
	    strlen(s->st_MsgId) == idLen &&
	    strncmp(msgId, s->st_MsgId, idLen) == 0
	) {
	    if (DebugOpt)
		printf(" found\n");
	    return(s);
	}
    }
    if (DebugOpt)
	printf(" not found\n");
    return(NULL);
}

/*
 * Refile pending streams
 */

void
RefilePendingStreams(FILE *fo)
{
    int i;

    for (i = 0; i < MAXSTREAM; ++i) {
	Stream *s = &StreamAry[i];

	if (s->st_State != STATE_EMPTY) {
	    fprintf(fo, "%s\t%s\n", s->st_RelPath, s->st_MsgId);
	    free(s->st_MsgId);
	    free(s->st_RelPath);
	    s->st_MsgId = NULL;
	    s->st_RelPath = NULL;
	    s->st_State = STATE_EMPTY;
	}
    }
}

/*
 * Misc subroutines
 */

void 
logit(const char *ctl, ...)
{
    va_list va;
    char buf[1024];

    sprintf(buf, "%s:%s ", 
	((HostName) ? HostName : "<unknown>"),
	((CurrentBatchFile) ? CurrentBatchFile : "<unknown>")
    );
    va_start(va, ctl);
    vsprintf(buf + strlen(buf), ctl, va);
    va_end(va);
    syslog(LOG_NOTICE, "%s", buf);
    if (DebugOpt)
	printf("%s%s", buf, ((buf[0] && buf[strlen(buf)-1] == '\n') ? "" : "\n"));
}


void 
logStats(const char *description)
{
    int secs = time(NULL) - DeltaStart;

    /*
     * If we got something through, cut off LastErrBuf
     */

    if (Count)
	LastErrBuf[9] = 0;

    credtime(OUR_DELAY);

    /*
     * Log!
     */

    logit("%s secs=%-4d acc=%-4d dup=%-4d rej=%-4d tot=%-4d (%d/min %d/%d mS) %s\n",
	description,
	secs,
	AcceptedCnt,
	RefusedCnt,
	RejectedCnt,
	Count,
	(secs) ? Count * 60 / secs : 0,
	((MsCount) ? OurMs / MsCount : -1),
	((MsCount) ?  TheirMs / MsCount : -1),
	LastErrBuf
    );
    DeltaStart = time(NULL);
    AcceptedCnt = 0;
    RefusedCnt = 0;
    RejectedCnt = 0;
    Count = 0;

    MsCountTotal += MsCount;
    OurMsTotal += OurMs;
    TheirMsTotal += TheirMs;

    MsCount = 0;
    OurMs = 0;
    TheirMs = 0;
}

/*
 * commandResponse() optionally send command, optionally read a response
 */

static char Buf[4096];
static int Bufb;
static int Bufe;

void
clearResponseBuf(void)
{
    Bufb = Bufe = 0;
}

int
commandResponse(int cfd, char **rptr, const char *ctl, ...)
{
    va_list va;
    int r = ERR_UNKNOWN;
    int n;

    if (ctl) {
	char tmp[4096];

	va_start(va, ctl);
	vsprintf(tmp, ctl, va);
	va_end(va);
	credtime(OUR_DELAY);
	write(cfd, tmp, strlen(tmp));
	credtime(THEIR_DELAY);

	if (DebugOpt) {
	    int i;

	    printf(">> ");
	    for (i = 0; tmp[i]; ++i) {
		if (isprint(tmp[i]))
		    printf("%c", tmp[i]);
		else
		    printf("[%02x]", tmp[i]);
	    }
	    puts("");
	}
    }
    if (rptr) {
	if (Timeout > 0)
	    alarm(Timeout);

	*rptr = NULL;

	/*
	 * Find next complete line in buffer, read more data if
	 * necessary.
	 */

	n = 0;

	credtime(OUR_DELAY);

	do {
	    int i;

	    credtime(THEIR_DELAY);
	    Bufe += n;
	    for (i = Bufb; i < Bufe; ++i) {
		if (Buf[i] == '\n') {
		    Buf[i] = 0;
		    if (i != Bufb && Buf[i-1] == '\r')
			Buf[i-1] = 0;
		    *rptr = Buf + Bufb;
		    Bufb = i + 1;
		    r = strtol(*rptr, NULL, 10);
		    break;
		}
	    }
	    if (i != Bufe)
		break;
	    /*
	     * We may have run out of space in the
	     * buffer, try to left-justify the data
	     * to make more room.  If we can't, the line
	     * is just too long.
	     */
	    if (Bufe == sizeof(Buf) - 1) {
		if (Bufb == 0)
		    break;
		memcpy(Buf, Buf + Bufb, Bufe - Bufb);
		Bufe -= Bufb;
		Bufb = 0;
	    }
	    credtime(OUR_DELAY);
	} while ((n = read(cfd, Buf + Bufe, sizeof(Buf) - Bufe - 1)) > 0);
	credtime(THEIR_DELAY);
	++MsCount;
	if (Timeout > 0)
	    alarm(0);
	if (DebugOpt)
	    printf("<< %s\n", (*rptr) ? *rptr : "<interrupt/error>");
	if (*rptr) {
	    strncpy(LastErrBuf, *rptr, 32);
	    LastErrBuf[32] = 0;
	} else {
	    strcpy(LastErrBuf, "<interrupt/error>");
	}
    } else {
	r = 0;
    }
    return(r);
}

void
sigTerm(int sigNo)
{
    TermFlag = 1;
    if (sigNo == SIGALRM) {
	if (KillFd >= 0)
	    close(KillFd);
    }
}

void 
bsprintf(const char *ctl, ...)
{
    if (StatBuf) {
	va_list va;
	int i;

	StatBuf[0] = 0;
 	if (MsCount > 4)
	    sprintf(StatBuf, "%d/%dmS ", OurMs / MsCount, TheirMs / MsCount);
	i = strlen(StatBuf);
	va_start(va, ctl);
	i += vsprintf(StatBuf + i, ctl, va);
	va_end(va);
	memset(StatBuf + i, ' ', StatSize - i - 1);
    }
}

void
credtime(int whos)
{
    struct timeval tv;
    static struct timeval Tv;

    gettimeofday(&tv, NULL);

    if (whos) {
	int ms = (tv.tv_usec + 1000000 - Tv.tv_usec) / 1000 +
	     (tv.tv_sec - Tv.tv_sec - 1) * 1000;

	switch(whos) {
	case OUR_DELAY:
	    OurMs += ms;
	    break;
	case THEIR_DELAY:
	    TheirMs += ms;
	    break;
	}
    }
    Tv = tv;
}

void
credreset(void)
{
    OurMs = TheirMs = 0;
}

typedef struct {
    char rb_Buf[MAXCLINE-sizeof(int)*3];
    int  rb_Base;
    int  rb_Index;
    int	 rb_Len;
} RBlock;

RBlock *RBlockAry[MAXFILEDES];

int
readline(int fd, char *buf, int size)
{
    RBlock *rbs;

    if (fd < 0 || fd >= MAXFILEDES) {
	logit("readline: bad fd: %d\n", fd);
	return(-1);
    }
    if ((rbs = RBlockAry[fd]) == NULL) {
	rbs = calloc(sizeof(RBlock), 1);
	RBlockAry[fd] = rbs;
    }
    
    /*
     * Ensure there is enough room for a full line
     */

    if (rbs->rb_Base > 0 && rbs->rb_Len > MAXCLINE / 2) {
	memmove(rbs->rb_Buf, rbs->rb_Buf + rbs->rb_Base, rbs->rb_Len - rbs->rb_Base);
	rbs->rb_Len -= rbs->rb_Base;
	rbs->rb_Index -= rbs->rb_Base;
	rbs->rb_Base = 0;
    }

    for (;;) {
	int n;

	/*
	 * Look for newline
	 */

	for (n = rbs->rb_Index; n < rbs->rb_Len; ++n) {
	    if (rbs->rb_Buf[n] == '\n') {
		int s = ++n - rbs->rb_Base;

		if (s >= size)
		    s = size - 1;
		memmove(buf, rbs->rb_Buf + rbs->rb_Base, s);
		buf[s] = 0;

		if (n == rbs->rb_Len) {
		    rbs->rb_Base = 0;
		    rbs->rb_Index = 0;
		    rbs->rb_Len = 0;
		} else {
		    rbs->rb_Base = n;
		    rbs->rb_Index = n;
		}
		return(0);
	    }
	}
	rbs->rb_Index = n;

	/*
	 * None found, read and loop
	 */

	if (rbs->rb_Len == sizeof(rbs->rb_Buf)) {
	    logit("Line too long in batchfile!\n");
	    return(-1);
	}
	n = read(fd, rbs->rb_Buf + rbs->rb_Len, sizeof(rbs->rb_Buf) - rbs->rb_Len);
	if (n <= 0)
	    break;
	rbs->rb_Len += n;
    }
    return(-1);
}

int
ValidMsgId(char *msgid)
{
    int l = strlen(msgid);

    if (msgid[0] == '<' && msgid[l-1] == '>')
	return(1);
    return(0);
}

